// Mandelbrot sample
// submitted by Mark Granger, NewTek

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <cuda_runtime_api.h>
#include <cuda_gl_interop.h>
#include "cutil.h"
#include "Mandelbrot_kernel.h"
#include "Mandelbrot_gold.h"


/////////////////////////////
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
/////////////////////////////
//
//

/**/
int sanity=0;
int portNumber;
int sockfd;
/**/

// Set to 1 to run on the CPU instead of the GPU for timing comparison.
#define RUN_CPU 0

// Set to 1 to time frame generation
#define RUN_TIMING 0

// Random number macros
#define RANDOMSEED(seed) ((seed) = ((seed) * 1103515245 + 12345))
#define RANDOMBITS(seed, bits) ((unsigned int)RANDOMSEED(seed) >> (32 - (bits)))

// Define types of requests
#define GETNEWFRAME 0x0
#define ZOOMIN      0x2
#define ZOOMOUT     0x3
#define MOVERIGHT   0x4
#define MOVEUP      0x5
#define MOVELEFT    0x6
#define MOVEDOWN    0x7
#define TERMINATE   0x8
#define MOUSE       0x9
#define CLICK       0x1
#define MOTION      0x2
#define LEFTCLICK   0x0
#define RIGHTCLICK  0x1

//OpenGL PBO and texture "names"
GLuint gl_PBO, gl_Tex;

//Mandelbrot data
uchar4 * mandelbrot_data;
uchar4 * gpu_mandelbrot_data;

//Original image width and height
int imageW, imageH;

// Starting iteration limit
int crunch = 512;

// Starting position and scale
double xOff = -0.5;
double yOff = 0.0;
double scale = 3.2;

// Starting stationary position and scale motion
double xdOff = 0.0;
double ydOff = 0.0;
double dscale = 1.0;

// Starting animation frame and anti-aliasing pass 
int animationFrame = 0;
int animationStep = 0;
int pass = 0;

// Starting color multipliers and random seed
int colorSeed = 0;
uchar4 colors;

// Timer ID
unsigned int hTimer;

// User interface variables
int lastx = 0;
int lasty = 0;
bool leftClicked = false;
bool rightClicked = false;

// Environment Variables
int rec_width, rec_heigh, max_pass = 8;

#define BUFFER_DATA(i) ((char *)0 + i)

// Get a sub-pixel sample location
void GetSample(int sampleIndex, float &x, float &y)
{    
    static const unsigned char pairData[128][2] = {
        { 64,  64}, {  0,   0}, {  1,  63}, { 63,   1}, { 96,  32}, { 97,  95}, { 36,  96}, { 30,  31},
        { 95, 127}, {  4,  97}, { 33,  62}, { 62,  33}, { 31, 126}, { 67,  99}, { 99,  65}, {  2,  34},
        { 81,  49}, { 19,  80}, {113,  17}, {112, 112}, { 80,  16}, {115,  81}, { 46,  15}, { 82,  79},
        { 48,  78}, { 16,  14}, { 49, 113}, {114,  48}, { 45,  45}, { 18,  47}, { 20, 109}, { 79, 115},
        { 65,  82}, { 52,  94}, { 15, 124}, { 94, 111}, { 61,  18}, { 47,  30}, { 83, 100}, { 98,  50},
        {110,   2}, {117,  98}, { 50,  59}, { 77,  35}, {  3, 114}, {  5,  77}, { 17,  66}, { 32,  13},
        {127,  20}, { 34,  76}, { 35, 110}, {100,  12}, {116,  67}, { 66,  46}, { 14,  28}, { 23,  93},
        {102,  83}, { 86,  61}, { 44, 125}, { 76,   3}, {109,  36}, {  6,  51}, { 75,  89}, { 91,  21},
        { 60, 117}, { 29,  43}, {119,  29}, { 74,  70}, {126,  87}, { 93,  75}, { 71,  24}, {106, 102},
        {108,  58}, { 89,   9}, {103,  23}, { 72,  56}, {120,   8}, { 88,  40}, { 11,  88}, {104, 120},
        { 57, 105}, {118, 122}, { 53,   6}, {125,  44}, { 43,  68}, { 58,  73}, { 24,  22}, { 22,   5},
        { 40,  86}, {122, 108}, { 87,  90}, { 56,  42}, { 70, 121}, {  8,   7}, { 37,  52}, { 25,  55},
        { 69,  11}, { 10, 106}, { 12,  38}, { 26,  69}, { 27, 116}, { 38,  25}, { 59,  54}, {107,  72},
        {121,  57}, { 39,  37}, { 73, 107}, { 85, 123}, { 28, 103}, {123,  74}, { 55,  85}, {101,  41},
        { 42, 104}, { 84,  27}, {111,  91}, {  9,  19}, { 21,  39}, { 90,  53}, { 41,  60}, { 54,  26},
        { 92, 119}, { 51,  71}, {124, 101}, { 68,  92}, { 78,  10}, { 13, 118}, {  7,  84}, {105,   4}
    };

    x = (1.0f / 128.0f) * (0.5f + (float)pairData[sampleIndex][0]);
    y = (1.0f / 128.0f) * (0.5f + (float)pairData[sampleIndex][1]);
} // GetSample

// OpenGL update function
void update(void)
{
//    uchar4 *d_dst = NULL;
    if ((xdOff != 0.0) || (ydOff != 0.0)) {
        xOff += xdOff;
        yOff += ydOff;
        pass = 0;
    }
    if (dscale != 1.0) {
        scale *= dscale;
        pass = 0;
    }
        
    do {
	    // Render anti-aliasing passes until we run out time (60fps approximately)
	    float xs, ys;

	    // Get the anti-alias sub-pixel sample location
	    GetSample(pass & 127, xs, ys);

	    // Get the pixel scale and offset
	    double s = scale / (float)imageW;
	    double x = (xs - (double)imageW * 0.5f) * s + xOff;
	    double y = (ys - (double)imageH * 0.5f) * s + yOff;

	    // Run the mandelbrot generator
	    RunMandelbrot0(gpu_mandelbrot_data, imageW, imageH, crunch, x, y, s, colors, pass++, animationFrame);
	    cudaThreadSynchronize();
    } while(pass < max_pass);
    cudaMemcpy(mandelbrot_data, gpu_mandelbrot_data, imageW * imageH * 4 * sizeof(char), cudaMemcpyDeviceToHost);
} // displayFunc


/**********Network Functions*****************/
void waitForConnection(int port) {
   int mySocket; 
   if((mySocket=socket(PF_INET,SOCK_STREAM,0))==-1) {
      perror("socket()");
      exit(1);
   }

   int yes=1;
   if(setsockopt(mySocket,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int))==-1) {
      perror("setsockopt()");
      exit(1);
   }

   struct sockaddr_in my_addr;
   my_addr.sin_family=AF_INET;
   my_addr.sin_port=htons(port);
   my_addr.sin_addr.s_addr=INADDR_ANY;
   memset(my_addr.sin_zero, '\0', sizeof my_addr.sin_zero);

   if(bind(mySocket,(struct sockaddr*)&my_addr, sizeof(struct sockaddr))==-1) {
      perror("bind()");
      exit(1);
   }

   if(listen(mySocket,5)==-1) {
      perror("listen()");
      exit(1);
   }

   struct sockaddr_in remote_addr;
   int sin_size=sizeof remote_addr;
   
   fprintf(stdout,"Waiting for connection on port %d...\n",port);
   if((sockfd=accept(mySocket, (struct sockaddr*)&remote_addr, (socklen_t*)&sin_size))==-1) {
      perror("accept()");
      exit(1);
   }

   fprintf(stdout, "Accepted connection from %s\n",inet_ntoa(remote_addr.sin_addr));
}

void initEnvironmentVariables(void) {
   int env[3];
   recv(sockfd, env, 3*sizeof(int), 0);
   imageW = env[0];
   imageH = env[1];
   max_pass = env[2];
   printf("Received environment variables: imageW = %d, imageH = %d, max_pass = %d\n", imageW, imageH, max_pass);
}

void requestLoop() {

   char requestBuffer[16];
   int bytesRead,bytesSent;
   char request;

//   fprintf(stdout,"Waiting for requests...\n");
   while(true) {
      bytesRead=recv(sockfd, (void*)requestBuffer,16 * sizeof(char),0);
      if(bytesRead==-1) {
         perror("recv()");
         exit(0);
      }
      if(bytesRead==0) exit(0);

      request=requestBuffer[0];

      if(request == GETNEWFRAME) {
         fprintf(stderr,"New frame requested\n");
	 update( );
         bytesSent=send(sockfd,(void*)mandelbrot_data, imageW * imageH * 4 * sizeof(char),0);
         if(bytesSent==-1) {perror("send()"); exit(1);}
//         fprintf(stderr,"Sent %d bytes\n",bytesSent);
      }
      else if(request==ZOOMIN) {
//         fprintf(stdout, "Received request to zoom in\n");
         scale /= 1.1;
	 pass = 0;
      }
      else if(request==ZOOMOUT) {
//         fprintf(stdout, "Received request to zoom out\n");
         scale *= 1.1;
	 pass = 0;
      }
      else if(request==MOVELEFT) {
//         fprintf(stdout, "Received request to move left\n");
         xOff -= 0.05f * scale;
	 pass = 0;
      }
      else if(request==MOVERIGHT) {
//         fprintf(stdout, "Received request to move right\n");
         xOff += 0.05f * scale;
	 pass = 0;
      }
      else if(request==MOVEUP) {
//         fprintf(stdout, "Received request to move up\n");
         yOff += 0.05f * scale;
	 pass = 0;
      }
      else if(request==MOVEDOWN) {
//         fprintf(stdout, "Received request to move down\n");
         yOff -= 0.05f * scale;
	 pass = 0;
      }
      else if(request==MOUSE) {
//         fprintf(stdout, "Received mouse event\n");
	 if(requestBuffer[1] == MOTION) {
	    int x_y[2];
	    memcpy(x_y, &requestBuffer[2], 2*sizeof(int));

	    double fx = (double)((x_y[0] - lastx) / 10) / (double)(imageW);
	    double fy = (double)((x_y[1] - lasty) / 10) / (double)(imageH);

	    if(leftClicked) {
	        xdOff = fx * scale;
		ydOff = fy * scale;
	    }
	    else {
		xdOff = 0.0f;
		ydOff = 0.0f;
	    }

	    if(rightClicked) {
		if(fy > 0.0f) {
		    dscale = 1.0 - fy;
		    dscale = dscale < 1.05 ? dscale : 1.05f;
		}
		else {
		    dscale = 1.0 / (1.0 + fy);
		    dscale = dscale > (1.0 / 1.05) ? dscale : (1.0 / 1.05);
		}
	    }
	    else {
		dscale = 1.0;
	    }

//	    fprintf(stdout, "Received a motion event at %d, %d\n", x_y[0], x_y[1]); 
	 }
	 else if(requestBuffer[1] == CLICK) {
	    if(requestBuffer[2] == LEFTCLICK) {
	    	leftClicked = !leftClicked;
	    }
	    else if(requestBuffer[2] == RIGHTCLICK) {
		rightClicked = !rightClicked;
	    }
	    int x_y[2];
	    memcpy(x_y, &requestBuffer[3], 2*sizeof(int));

	    lastx = x_y[0];
	    lasty = x_y[1];

	    xdOff = 0.0;
	    ydOff = 0.0;
	    dscale = 1.0;
//	    fprintf(stdout, "Received a click event %d, %d, %d\n", requestBuffer[2], x_y[0], x_y[1]); 
	 }
	 else {
	    fprintf(stdout, "Received an incorrect mouse request\n"); 
	 }
      }
      else if(request==TERMINATE) {
         fprintf(stdout,"Received request to terminate. Exiting...\n");
         exit(0);
      }
      else {
         fprintf(stderr,"Received invalid command %d\n",request);
         exit(1);
      }
   }
}


////////////////////////////////////////////////////////////////////////////////
// Main program
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    if(argc != 2) {
	printf("USAGE: %s <port_number>\n", argv[0]);
	exit(0);
    }
    waitForConnection(atoi(argv[1]));

    initEnvironmentVariables( );
    mandelbrot_data = (uchar4*) malloc(imageW * imageH * 4);
    cudaMalloc((void**)&gpu_mandelbrot_data, imageW * imageH * sizeof(uchar4));
    colors.w = 0;
    colors.x = 3;
    colors.y = 5;
    colors.z = 7;
//    printf("Data init done.\n");
    requestLoop( );
} // main
